unit Farmer;

interface

uses Windows, Classes, SyncObjs, Abstraction;

type
  TFarmer = class(TAFarmer)
    //Note that the farmer is expected to run in the primary thread
  protected
    FValues:PValue;     //First of values, which are stored in the memory
    FCommitted:PValue;  //A value to be assigned next to a worker
    FLimit:PValue;      //Last of the stored values
    FSum:TValue;        //The so-far known sum of all values

    FDataGuard,         //Guarding access to FCommitted and FSum
    FMessageGuard:TCriticalSection; //Guarding access to messages

    FWait,                //Manual-reset event signalizing all workers are done
    FMessageAdded:THandle;//Manual-reset event, new message to be displayed
    FWorkerCount:integer; //Number of registered threads (threads to wait for)

    FMessages:TStringList;//Messages to be displayed

    procedure GenerateValues(ASize:integer); dynamic;
        //Generates values to be summed up
  public
    constructor Create(ACount:integer);
    destructor Destroy; override;

    function AskForValues(out Values:PValue; out ALimit:PValue):boolean; override;
    procedure ReportSum(ASum:TValue); override;
    procedure LogMessage(const AMessage:String); override; 
    procedure AddWorkerRef; override;
    procedure ReleaseWorker; override;

    procedure WaitForWorkers;
    property Sum:TValue read FSum;
  end;

implementation

uses SysUtils, Strings;

constructor TFarmer.Create(ACount:integer);
begin
  //Allocate the memory
  GetMem(FValues, sizeof(TValue)*ACount);
  FLimit:=PValue(integer(FValues)+sizeof(TValue)*(ACount-1));
  FCommitted:=FValues;

  //Init what's necessary
  GenerateValues(ACount);
  FSum:=0;
  FWorkerCount:=0;

  //Create events
  FWait:=CreateEvent(nil, true, false, nil);
  FMessageAdded:=CreateEvent(nil, true, false, nil);

  //critical sections
  FDataGuard:=TCriticalSection.Create;
  FMessageGuard:=TCriticalSection.Create;

  //and the message holder as well
  FMessages:=TStringList.Create;
end;

destructor TFarmer.Destroy;
begin
  FDataGuard.Free;
  FMessageGuard.Free;

  CloseHandle(FWait);
  CloseHandle(FMessageAdded);

  FreeMem(FValues);
  FMessages.Free;

  inherited;
end;

procedure TFarmer.GenerateValues(ASize:integer);
var i:integer;
   index:PValue;
begin
  index:=FValues;
  for i:=1 to ASize do
    begin
      index^:=i;
      inc(index);
    end;
end;

function TFarmer.AskForValues(out Values:PValue; out ALimit:PValue):boolean;
var tmp:integer; 
begin
  FDataGuard.Acquire;
  try
    //Are there some values left for the calling worker?
    result:=integer(FCommitted)<=integer(FLimit);
    if result then
       begin
          Values:=FCommitted;
          inc(FCommitted, WorkSegmentation);
            //FComitted points to next number to be assigned next time

          ALimit:=FCommitted;
          dec(ALimit);
          if integer(ALimit)>integer(FLimit) then ALimit:=FLimit;//overflow check
          //ALimit points to the last number of assigned values

          tmp:=integer(FLimit)-integer(ALimit);
          LogMessage(Format(rsFarmerProgress, [tmp]));
        end;
  finally
    FDataGuard.Release;
  end;
end;

procedure TFarmer.ReportSum(ASum:TValue); 
begin
  FDataGuard.Acquire;
  try
    FSum:=FSum+ASum;
  finally
    FDataGuard.Release;
  end;
end;

procedure TFarmer.WaitForWorkers;

procedure DumpMessages;
var i:integer;
begin
  //Just write all messages we have buffered
  FMessageGuard.Acquire;
  try
    for i:=0 to FMessages.Count - 1 do
      writeln(FMessages[i]);
    FMessages.Clear;

    ResetEvent(FMessageAdded);  //one excessive reset only    
  finally
    FMessageGuard.Release;
  end;
end;

const
  hiWait = 0;
  hiMsg  = 1;

var Handles:array[0..1] of THandle;
    res:integer;

begin
  //We are the main thread, and as a such one, we will serialize
  //the output of messages to the standard output

  Handles[hiWait]:=FWait;
  Handles[hiMsg]:=FMessageAdded;

  repeat
    res:=WaitForMultipleObjects(Length(Handles), @Handles, False, INFINITE);
      //The code gets here as soon as all workers are done,
      //or there is a new message to be displayed. Because
      //of thread switching, several messages may get buffered
      //prior getting here.
    DumpMessages;
    
  until res=WAIT_OBJECT_0+hiWait;
end;

procedure TFarmer.AddWorkerRef;
begin
  InterlockedIncrement(FWorkerCount);
end;

procedure TFarmer.ReleaseWorker;
begin
  InterlockedDecrement(FWorkerCount);
  if FWorkerCount<1 then SetEvent(FWait);
end;

procedure TFarmer.LogMessage(const AMessage:String);
begin
  FMessageGuard.Acquire;
  try
    FMessages.Add(AMessage);
    SetEvent(FMessageAdded);
  finally
    FMessageGuard.Release;
  end;
end;

initialization
  Randomize;
end.
